# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging

import lightning.pytorch as pl
import torch.nn as nn
import torch.utils.data
import torchvision.models as models
import torchvision.transforms

from ..data.typing import DataLoader

logger = logging.getLogger(__name__)


class Alexnet(pl.LightningModule):
    """Alexnet module.

    Note: only usable with a normalized dataset
    """

    def __init__(
        self,
        criterion=None,
        criterion_valid=None,
        optimizer=None,
        optimizer_configs=None,
        pretrained=False,
        augmentation_transforms=[],
    ):
        super().__init__()

        self.name = "alexnet"

        self.augmentation_transforms = torchvision.transforms.Compose(
            augmentation_transforms
        )

        self.criterion = criterion
        self.criterion_valid = criterion_valid

        self.optimizer = optimizer
        self.optimizer_configs = optimizer_configs

        self.normalizer = None
        self.pretrained = pretrained

        # Load pretrained model
        if not pretrained:
            weights = None
        else:
            logger.info(f"Loading pretrained {self.name} model weights")
            weights = models.AlexNet_Weights.DEFAULT

        self.model_ft = models.alexnet(weights=weights)

        # Adapt output features
        self.model_ft.classifier[4] = nn.Linear(4096, 512)
        self.model_ft.classifier[6] = nn.Linear(512, 1)

    def forward(self, x):
        x = self.normalizer(x)
        x = self.model_ft(x)

        return x

    def set_normalizer(self, dataloader: torch.utils.data.DataLoader) -> None:
        """Initializes the normalizer for the current model.

        This function is NOOP if ``pretrained = True`` (normalizer set to
        imagenet weights, during contruction).

        Parameters
        ----------

        dataloader: :py:class:`torch.utils.data.DataLoader`
            A torch Dataloader from which to compute the mean and std.
            Will not be used if the model is pretrained.
        """
        if self.pretrained:
            from .normalizer import make_imagenet_normalizer

            logger.warning(
                f"ImageNet pre-trained {self.name} model - NOT "
                f"computing z-norm factors from train dataloader. "
                f"Using preset factors from torchvision."
            )
            self.normalizer = make_imagenet_normalizer()
        else:
            from .normalizer import make_z_normalizer

            logger.info(
                f"Uninitialised {self.name} model - "
                f"computing z-norm factors from train dataloader."
            )
            self.normalizer = make_z_normalizer(dataloader)

    def balance_losses_by_class(
        self, train_dataloader: DataLoader, valid_dataloader: DataLoader
    ):
        """Reweights loss weights if possible.

        Parameters
        ----------

        train_dataloader
            The data loader to use for training

        valid_dataloader
            The data loader to use for validation


        Raises
        ------

        RuntimeError
            If train or validation losses are not of type
            :py:class:`torch.nn.BCEWithLogitsLoss`.
        """
        from .loss_weights import get_label_weights

        if isinstance(self.criterion, torch.nn.BCEWithLogitsLoss):
            logger.info("Reweighting BCEWithLogitsLoss training criterion.")
            weights = get_label_weights(train_dataloader)
            self.criterion = torch.nn.BCEWithLogitsLoss(pos_weight=weights)
        else:
            raise RuntimeError(
                "Training loss is not BCEWithLogitsLoss - dunno how to balance"
            )

        if isinstance(self.criterion_valid, torch.nn.BCEWithLogitsLoss):
            logger.info("Reweighting BCEWithLogitsLoss validation criterion.")
            weights = get_label_weights(valid_dataloader)
            self.criterion_valid = torch.nn.BCEWithLogitsLoss(weights)
        else:
            raise RuntimeError(
                "Validation loss is not BCEWithLogitsLoss - dunno how to balance"
            )

    def training_step(self, batch, batch_idx):
        images = batch[0]
        labels = batch[1]["label"]

        # Increase label dimension if too low
        # Allows single and multiclass usage
        if labels.ndim == 1:
            labels = torch.reshape(labels, (labels.shape[0], 1))

        # Forward pass on the network
        augmented_images = [
            self.augmentation_transforms(img).to(self.device) for img in images
        ]
        # Combine list of augmented images back into a tensor
        augmented_images = torch.cat(augmented_images, 0).view(images.shape)
        outputs = self(augmented_images)

        training_loss = self.criterion(outputs, labels.float())

        return {"loss": training_loss}

    def validation_step(self, batch, batch_idx, dataloader_idx=0):
        images = batch[0]
        labels = batch[1]["label"]

        # Increase label dimension if too low
        # Allows single and multiclass usage
        if labels.ndim == 1:
            labels = torch.reshape(labels, (labels.shape[0], 1))

        # data forwarding on the existing network
        outputs = self(images)

        validation_loss = self.criterion_valid(outputs, labels.float())

        if dataloader_idx == 0:
            return {"validation_loss": validation_loss}
        else:
            return {f"extra_validation_loss_{dataloader_idx}": validation_loss}

    def predict_step(self, batch, batch_idx, dataloader_idx=0, grad_cams=False):
        images = batch[0]
        labels = batch[1]["label"]
        names = batch[1]["name"]

        outputs = self(images)
        probabilities = torch.sigmoid(outputs)

        return names[0], torch.flatten(probabilities), torch.flatten(labels)

    def configure_optimizers(self):
        optimizer = self.optimizer(self.parameters(), **self.optimizer_configs)
        return optimizer
